/*
 * Common CAMAC definitions
 */

#ifndef _LINUX_CAMAC_H
#define _LINUX_CAMAC_H

#ifndef __KERNEL__
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#endif /* ndef __KERNEL__ */

/*
 * Define the ioctl numbers 
 */

#define CAMAC_TEST_CRATE 33      /* tests the aviability of the crate */
#define CAMAC_STATUS 34          /* read X and Q flags */
#define CAMAC_BLOCK_MODE 35      /* Block transfer mode */
#define CAMAC_SINGLE_MODE 36     /* 16/24 bit transfer mode, default */
#define CAMAC_QSTOP_MODE 37      /* Q-Stop tarnsfer mode */
#define CAMAC_QREPEAT_MODE 38    /* Q-Repeat transfer mode */
#define CAMAC_ADDRSCAN_MODE 39   /* Address scan transfer mode */
#define CAMAC_ALLOW_BROADCAST 40 /* enable broadcasting for this crate */
#define CAMAC_DENY_BROADCAST 41  /* disable broadcasting for this crate */
#define CAMAC_C_Z_S2 42          /* make c-z-s2 cycle */
#define CAMAC_INITIALIZE 43      /* Camac Z-cycle */
#define CAMAC_CLEAR 44           /* Camac C-cycle */
#define CAMAC_WAIT_LAM 45        /* wait for a LAM if timeout set */
#define CAMAC_NON_DATA 46        /* Camac cycle without data */
#define CAMAC_SET_INHIBIT 47     /* set CAMAC inhibit flag */
#define CAMAC_CLEAR_INHIBIT 48   /* clear CAMAC inhibit flag */

/*
 * Do not change the following numbers, they are hard-coded
 * to the according CAMAC functions
 */

#define CAMAC_TEST_LAM 8        /* test look-at-me flags */
#define CAMAC_CLEAR_1 9         /* clear group 1 register */
#define CAMAC_CLEAR_LAM 10      /* clear look-at-me flags */
#define CAMAC_CLEAR_2 11        /* clear group 2 register */
#define CAMAC_DISABLE 24        /* disable command */
#define CAMAC_EXECUTE 25        /* execute command */
#define CAMAC_ENABLE 26         /* enable command */
#define CAMAC_TEST_STATUS 27    /* test status command */

/*
 * The CAMAC read and write functions
 */

#define CAMAC_READ_1 0          /* read group 1 register */
#define CAMAC_READ_2 1          /* read group 2 register */
#define CAMAC_READ_CLEAR_1 2    /* read and clear group 1 register */
#define CAMAC_CPL_READ_1 3      /* read complement of group 1 register */
#define CAMAC_WRITE_1 16        /* write group 1 register */
#define CAMAC_WRITE_2 17        /* write group 2 register */
#define CAMAC_SEL_WRITE_1 18    /* selective write group 1 register */
#define CAMAC_SEL_WRITE_2 19    /* selective write group 2 register */

/*
 * Define help macros for the N, A, F code
 */

#define NAF(n,a,f) ((((a) & 0x0f) << 10) | (((f) & 0x1f) << 5) | ((n) & 0x1f))
#define N_NAF(naf) ((naf) & 0x1f)
#define A_NAF(naf) (((naf) >> 10) & 0x0f)
#define F_NAF(naf) (((naf) >> 5) & 0x1f)

/*
 * Define macros for getting the X and Q flags
 * the LAM flag is only usable for newer Boards
 */

#define CAMAC_Q_FLAG(qx) ((qx) & 0x01)
#define CAMAC_X_FLAG(qx) (((qx) >> 1) & 0x01)
#define CAMAC_LAM_FLAG(qx) (((qx) >> 2) & 0x01)

/*
 * The debugging flags
 */

#define CAMAC_DBG_ERROR 0x01 /* only error messages */
#define CAMAC_DBG_OPEN 0x02  /* open/close devices */
#define CAMAC_DBG_RW 0x04    /* lseek/read/write operations */
#define CAMAC_DBG_IOCTL 0x08 /* ioctl operations */
#define CAMAC_DBG_IO 0x10    /* all i/o operations (not implemented) */

/*
 * Generic CAMAC commands are just inline functions
 */ 
  
#ifndef __KERNEL__

__inline__ int CNAF(int crate,int N,int A,int F,void *data)
{
  int CNAF_f = F;
  int qx;

  lseek(crate, NAF(N,A,CNAF_f), SEEK_SET);
  if ((CNAF_f & 0x18) == 0)
    read(crate, data, 2);
  else if ((CNAF_f & 0x18) == 0x10)
    write(crate, data, 2);
  else
    ioctl(crate, CAMAC_NON_DATA, data);

  ioctl(crate, CAMAC_STATUS, &qx);
  return qx;
}
                
__inline__ int CNAF24(int crate,int N,int A,int F,void *data)
{
  int CNAF_f = F;
  int qx;

  lseek(crate, NAF(N,A,CNAF_f), SEEK_SET);
  if ((CNAF_f & 0x18) == 0)
    read(crate, data, 4);
  else if ((CNAF_f & 0x18) == 0x10)
    write(crate, data, 4);
  else
    ioctl(crate, CAMAC_NON_DATA, data);

  ioctl(crate, CAMAC_STATUS, &qx);
  return qx;
}
                
#define CNAF16 CNAF
#define CNAF32 CNAF24
#endif /* ndef __KERNEL__ */

#endif /* _LINUX_CAMAC_H */

