/*
 * This is a driver for the Aachen DBCC CAMAC controller (Zeuthen version)
 * Author: Ole Streicher <ole@ifh.de>
 * I hold the full copyright of this software (excl. the hardware part,
 * which was stolen from Thomas Mikolajskis ACCAMAC library).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Acknowledgements:
 *  Thomas Mikolajski, for his ACCAMAC library and for extensive testing
 *  ...
 */

/*
 * There are two possibilities to use this driver:
 * 1. directly compiled-in
 *   Copy the file dbcc.c to linux/drivers/char/, dbcc.h and camac.h to 
 *   linux/include/linux/ and install the patch camac.patch. Then recompile 
 *   your kernel, answering the question for a CAMAC driver with "yes".
 *   At boot time, you are able to give the base address of the interface 
 *   card with the option "dbcc=0x320", or to disable the driver with the 
 *   option "dbcc=off". This method is no lonmger recommended!
 * 2. as a loadable module
 *   I recommend this option because it is easier to install and to handle.
 *   Type "make" to compile the module, and "make install" to install it.
 *   The command "insmod dbcc.o" will install the driver, the option 
 *   "io=0x320" will change the Base address, "major=25" will change the 
 *   major number. Use "rmmod dbcc" to remove the module.
 */

/*
 * History:
 *  0.1 initial development version (ole 1.6.95)
 *  0.2 loadable modules and setup function added (ole 6.6.95)
 *  0.3 tested with a real crate (ole 7.6.95)
 *  0.4 start including AC_WAIT_LAM (ole 11.6.95)
 *  0.4.1 bug in ac_wait_lam(): we need to set the crate every time we
 *        read after scheduling (ole 14.6.95)
 *  0.4.2 change the memory allocation for block r/w to the common kmalloc()
 *        (ole 16.6.95)
 *  0.4.3 added the dbcc_enabled variable for builtin driver; now we can 
 *        disable the driver at LILO cmd line (ole 18.6.95)
 *  0.4.4 added the CAMAC_START_CYCLE to the ioctl Camac commands (9.8.95)
 *        ignore NULL pointer in the ioctl commands (9.8.95)
 *  0.4.5 ignore NULL pointer in the r/w commands (10.8.95)
 *  0.4.6 CAMAC_WAIT_LAM changed to CAMAC_SET_LAM, new CAMAC_WAIT_LAM waits 
 *         now directly for a LAM (10.8.95)
 *  0.4.7 added CAMAC_START_CYCLE to the CAMAC_WAIT_LAM ioctl command
 *        add a check the the user can really read/write the buffer (16.8.95)
 *  0.4.8 bug in r/w block mode: need to alloc count bytes instead of nwords
 *        start inserting debug code (24.8.95)
 *  0.4.9 bug in ioctl: insert dbcc_set() for CAMAC commands (29.9.95)
 *  0.4.10 open the crate non-exclusively in normal case (ole 16.11.95)
 *  0.5.0 start to include interrupt handling (17.9.95)
 *        first step: test irq capability at booting and open
 *        second step: rewrite waitfor_lam()
 *        collect all global Camac variables into the Camac structure
 *        new #defines in camac.h for r/w operations (ole 9.10.95)
 *        remove/change the check for the r/w functions (ole 9.10.95)
 *        include basic CAMAC functions into camac.h (ole 13.10.95)
 *        new ioctl: exec CAMAC function stored with lseek (ole 16.10.95)
 *        changed driver and file name to dbcc (ole 16.10.95)
 *        changed most ioctl names from AC_xxx to CAMAC_xxx (ole 16.10.95)
 *        the internal function names start now with "dbcc" (ole 16.11.95)
 *        added ioctl's and functions for C and Z cycle (ole 19.11.95)
 *        changed the behaviour of CAMAC_WAIT_LAM: now it waits directly
 *        for a LAM, read returns now immediately (ole 22.11.95)
 *        added ioctl names for Q-stop-mode, Q-Repeat-mode, 
 *        Address-scan-mode (ole 28.11.95)
 *        changed the CNAF macros to inline functions (ole 28.11.95)
 *  0.5.1 new ioctl()s: CAMAC_SET_INHIBIT, CAMAC_CLEAR_INHIBIT (ole 10.12.95)
 *        base address variable is no called "io" instead of "base" like
 *        in other modules (ole 28.12.95)
 *        write() prototype: char *buf --> const char *buf to avoid 
 *        warning with Linux1.3.x (ole 29.12.95)
 */

/*
 * Todo:
 * 1.Make driver more generic
 *  - check for the possibilities with the JORWAY controller
 *    * blockmode with automatic increment
 *    * blockmode until not Q response
 *    * 24 bit block modes?
 * 2.Other ideas
 *  - will we need a select() function?
 *  - the function CAMAC_WAIT_LAM should wait for any interrupt from the crate?
 *  - lseek() could allow also the directive SEEK_CUR which increases the 
 *    subaddress and then the position, keeping the function code
 * 3.Testing
 *  - test the functions:
 *    * write_blk
 *  - if we get an interrupt-capable board: use the interrupt
 *    * test the checks for IRQ-capable crates and boards
 *    * test the real wait-for-LAM code
 *  - test the non-exclusive opening of the crate
 *    * just test it
 *    * There could be a bug in dbcc_irq_handler(): it does not test for 
 *      the correct module. So, if 2 processes wait of different LAM in 
 *      the same crate, the first LAM will service both processes.
 * 4.Higher level stuff
 *  - write a library (C++ ?)
 *  - write a tk/tcl interface (for this, we need a simple HBOOK tk/tcl, too)
 *  - include all stuff into the Linux LAB project
 *  - insert all debugging stuff
 */

#include <linux/camac.h>
#include <linux/dbcc.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/segment.h>
#ifdef MODULE
#include <linux/module.h>
#ifndef UTS_RELEASE
#include <linux/version.h> 
#endif
static char kernel_version[] = UTS_RELEASE;
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#define CAMAC_VERSION "0.5.1"
#define STARTUPSTRING "CC16 CAMAC driver "CAMAC_VERSION" by Ole Streicher\n"

/* 
 * WAIT_IN_READ is obsolete code, waits for a LAM in the read() call.
 * Now (0.5.0) waiting for LAM is an extra ioctl() call.
 */

#undef WAIT_IN_READ

/* 
 * here come some static variables
 * the start address is tunable: insmod dbcc.o io=0x320 for loadable 
 * modules and dbcc=0x320 for built-in driver
 * Tunable for modules: insmod dbcc major=25
 * enable interrupt mode: insmod dbcc irq=5
 */

static unsigned long io = AC_BASE_ADDRESS;
#define BaseAddress io
#define CtrlInAddress (io + 0x00)
#define CtrlOutAddress (io + 0x00)
#define DataOutAddress (io + 0x02)
#define DataInAddress (io + 0x04)
static int major = AC_MAJOR;                          /* Major device number */
static int irq = 0;                             /* irq number, disabled if 0 */

static int currentDevice;
static int nOpen[AC_MAX_CRATES];     /* Number of open files for this device */
static void *ExclOpen[AC_MAX_CRATES];          /* Flag for exclusive opening */
static int Irq[AC_MAX_CRATES]; /* flag for interrupt capability of the crate */
static int Sleeping[AC_MAX_CRATES];   /* flag, if the crate waits for an int */
static struct wait_queue *wait_q[AC_MAX_CRATES];

/*
 * The operation modi of the device are stored for each opened device in
 * the file->private_data field, which is a pointer to the following 
 * structure
 */

typedef struct  {
  int Mode;                                                 /* transfer mode */
  int TimeOut;                                  /* timeout, in 10 msec units */
} cfg_t; /* Driver configuration, allocated in open() */

/*
 * The functions for the virtual file system
 */

static int lseek_dbcc(struct inode *, struct file *, off_t, int );
static int read_dbcc(struct inode *, struct file *, char *, int );
static int write_dbcc(struct inode *, struct file *, const char *, int );
static int ioctl_dbcc(struct inode *, struct file *, unsigned, unsigned long);
static int open_dbcc(struct inode *, struct file *);
static void close_dbcc(struct inode *, struct file *);

/*
 * The functions for loading and exiting the driver
 */

#ifdef MODULES
int init_module(void);
void cleanup_module(void);
#else
unsigned long dbcc_init(unsigned long);    /* called from drivers/char/mem.c */
void dbcc_setup(char *, int *);            /* called from init/main.c */
static int dbcc_enabled = 1;               /* set in setup, used init */
#endif

/*
 * The following functions are only internal used; they contain the
 * I/O code
 */

static int dbcc_test(void);              /* tests the actual crate */
static void dbcc_set(unsigned);          /* sets the crate nr */
static int dbcc_test_pcboard(void);      /* test the PC interface */
static int dbcc_test_irq(void);          /* test interrupt of the crate */
static void 
dbcc_irq_handler(int, struct pt_regs *); /* interrupt handler */
static void dbcc_enable_interrupt(void); /* enable interrupt */
static void dbcc_disable_interrupt(void);/* disable interrupt */
static void dbcc_allow_broadcast(void);  /* allow broadcasting */
static void dbcc_deny_broadcast(void);   /* deny broadcasting */
static int dbcc_status(void);            /* return X and Q status */
static int dbcc_ctrl(int);               /* CAMAC control functions */
static void dbcc_czs2(void);             /* make a c-z-s2 cycle */
static void dbcc_c_cycle(void);          /* make a C cycle */
static void dbcc_z_cycle(void);          /* make a Z cycle */
static void dbcc_set_inhibit(void);      /* set inhibit flag */
static void dbcc_clear_inhibit(void);    /* clear inhibit flag */
static int dbcc_wait_lam(int, int);      /* wait for LAM */
static int dbcc_wait_lam_irq(int, int);    /* (irq mode) */
static int dbcc_wait_lam_polling(int, int);/* (polling mode) */
static void dbcc_load_naf(int);          /* 1st part of CAMAC cycle */
static int dbcc_read_16(void);           /* 2nd part: read 16 bit */
static int dbcc_read_24(void);           /* 2nd part: read 24 bit */
static void dbcc_read_blk(short *, int); /* 2nd part: read in block mode */
static void dbcc_write_16(int);          /* 2nd part: write 16 bit */
static void dbcc_write_24(int);          /* 2nd part: write 24 bit */
static void dbcc_write_blk(short *, int);/* 2nd part: write in block mode */

/*
 * Debug stuff: a function which (kernel)prints, if the first parameter 
 * fits the debuglevel. If the AC_DEBUG flag isn't set (in the Makefile),
 * do nothing.
 */

#ifdef AC_DEBUG
static unsigned long debuglevel = 0L; /* default: no debug */
#define DEBUG(a,s...) if (a & debuglevel) printk(## s)
#else /* if !AC_DEBUG */
#define DEBUG(a,s...) /* do nothing */
#endif /* !AC_DEBUG */

/*
 * Here is the file_operation structure
 */

struct file_operations dbcc_fops = {
  lseek_dbcc,
  read_dbcc,
  write_dbcc,
  NULL,           /* readdir */
  NULL,           /* select */
  ioctl_dbcc,
  NULL,           /* mmap */
  open_dbcc,
  close_dbcc,
};


/*
 * The function lseek_dbcc() is used with the NAF macro as argument;
 * just origin 0 (from start) is allowed
 */

static int 
lseek_dbcc(struct inode * inode, struct file * file, off_t offset, int orig)
{
  if (orig == 0) {
    DEBUG(CAMAC_DBG_RW,"camac: lseek(c=%u, n=%li, a=%li, f=%li)\n",
	  MINOR(inode->i_rdev), N_NAF(offset),A_NAF(offset),F_NAF(offset));
    file->f_pos = offset;
    return file->f_pos;
  }
  else
    return -EINVAL;
} 

/*
 * The function read_dbcc() reads in single or block mode depending from
 * the value of Mode[]
 * In single mode, if one reads 1 or 2 bytes, a 16-bit-cycle is started
 * if one reads 3 or 4 bytes, a 24-bit-cycle is started
 * in block mode, We will only read an even number of bytes
 */

static int 
read_dbcc(struct inode * inode, struct file * file , char * buf, int count)
{

  int nwords = count/2;
  unsigned int crate = MINOR(inode->i_rdev);
  int result;

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  if (nwords > AC_MAX_WORDS)
    return -EINVAL;

  if (nwords == 0) 
    return 0;

  /*
   * Check that we can really write to the buffer
   */

  if ((result = verify_area(VERIFY_WRITE,buf,count)) && (buf != NULL))
    return result;

  /*
   * If the bits 3 and 4 of the function code are not 0, it is not a read 
   * operation
   * in this case we dont want to read anything
   */
#ifdef CHECK_RW_RANGE  
  if ((F_NAF(file->f_pos) & 0x18) != 0)
    return 0;
#endif /* CHECK_RW_RANGE */

  DEBUG(CAMAC_DBG_RW,"camac: read(crate=%u, n=%i, a=%i, f=%i, nw=%i)\n",
	crate, (int)N_NAF(file->f_pos),(int)A_NAF(file->f_pos),
	(int)F_NAF(file->f_pos), nwords);

  /*
   * look if the block transfer mode is on 
   */

  if (((cfg_t *)file->private_data)->Mode == CAMAC_BLOCK_MODE) {

    short *temp = kmalloc(count, GFP_USER);

    dbcc_set(crate);
#ifdef WAIT_IN_READ
    if (dbcc_wait_lam(file->f_pos, ((cfg_t *)file->private_data)->TimeOut))
      return -EIO;
#endif
    dbcc_load_naf(file->f_pos);
    dbcc_read_blk(temp, nwords);
    if (buf != NULL)
      memcpy_tofs(buf, temp, 2*nwords);
    kfree(temp);
    return 2*nwords;
  } 
  else { /* if single mode */

    int temp;

    switch (count) {
    case 1:
      dbcc_set(crate);
#ifdef WAIT_IN_READ
      if (dbcc_wait_lam(file->f_pos, ((cfg_t *)file->private_data)->TimeOut)) 
	return -EIO;
#endif
      dbcc_load_naf(file->f_pos | AC_START_CYCLE);
      temp = dbcc_read_16();
      if (buf != NULL)
       	put_fs_byte(temp, buf);
      return count;
    case 2:
      dbcc_set(crate);
#ifdef WAIT_IN_READ
      if (dbcc_wait_lam(file->f_pos, ((cfg_t *)file->private_data)->TimeOut)) 
	return -EIO;
#endif
      dbcc_load_naf(file->f_pos | AC_START_CYCLE);
      temp = dbcc_read_16();
      if (buf != NULL)
       	put_fs_word(temp, buf);
      return count;
    case 3:
      dbcc_set(crate);
#ifdef WAIT_IN_READ
      if (dbcc_wait_lam(file->f_pos, ((cfg_t *)file->private_data)->TimeOut)) 
	return -EIO;
#endif
      dbcc_load_naf(file->f_pos | AC_START_CYCLE);
      temp = dbcc_read_24();
      if (buf != NULL)
       	memcpy_tofs(buf, &temp, 3);
      return count;
    case 4:
      dbcc_set(crate);
#ifdef WAIT_IN_READ
      if (dbcc_wait_lam(file->f_pos, ((cfg_t *)file->private_data)->TimeOut)) 
	return -EIO;
#endif
      dbcc_load_naf(file->f_pos | AC_START_CYCLE);
      temp = dbcc_read_24();
      if (buf != NULL)
        put_fs_long(temp, buf);
      return count;
    default:
      return -EINVAL;
    } /* switch count */
  } /* if ! blockmode */

}

/*
 * The function write_dbcc() writes in single or block mode depending from
 * the value of Mode[]
 * In single mode, if one writes 1 or 2 bytes, a 16-bit-cycle is started
 * if one writes 3 or 4 bytes, a 24-bit-cycle is started
 * in block mode, We will only write an even number of bytes
 */

static int 
write_dbcc(struct inode * inode, struct file * file , const char * buf, int count)
{

  int nwords = count/2;
  unsigned int crate = MINOR(inode->i_rdev);
  int result;

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  if (nwords > AC_MAX_WORDS)
    return -EINVAL;

  if (nwords == 0) 
    return 0;

  /*
   * Check that we can really read from the buffer
   */

  if ((result = verify_area(VERIFY_READ,buf,count)))
    return result;
     
  /*
   * If the bits 3 and 4 of the function code are not 10, it is not a write
   * operation
   * in this case we dont want to write anything
   */
#ifdef CHECK_RW_RANGE  
  if ((F_NAF(file->f_pos) & 0x18) != 10)
    return 0;
#endif /* CHECK_RW_RANGE */

  DEBUG(CAMAC_DBG_RW,"camac: write(crate=%u, n=%i, a=%i, f=%i, nw=%i)\n",
	crate, (int)N_NAF(file->f_pos),(int)A_NAF(file->f_pos),
	(int)F_NAF(file->f_pos), nwords);

  /*
   * look if the block transfer mode is on 
   */
  
  if (((cfg_t *)file->private_data)->Mode == CAMAC_BLOCK_MODE) {

    short *temp = kmalloc(count, GFP_USER);

    memcpy_fromfs(temp, buf, 2*nwords);
    dbcc_set(crate);
    dbcc_load_naf(file->f_pos);
    dbcc_write_blk(temp, nwords);
    kfree(temp);
    return 2*nwords;
  } 
  else { /* if single mode */

    int temp;

    switch (count) {
    case 1:
      temp = get_fs_byte(buf);
      dbcc_set(crate);
      dbcc_load_naf(file->f_pos);
      dbcc_write_16(temp);
      return count;
    case 2:
      temp = get_fs_word(buf);
      dbcc_set(crate);
      dbcc_load_naf(file->f_pos);
      dbcc_write_16(temp);
      return count;
    case 3:
      memcpy_fromfs(&temp, buf, 3);
      dbcc_set(crate);
      dbcc_load_naf(file->f_pos);
      dbcc_write_24(temp);
      return count;
    case 4:
      temp = get_fs_long(buf);
      dbcc_set(crate);
      dbcc_load_naf(file->f_pos);
      dbcc_write_24(temp);
      return count;
    default:
      return -EINVAL;
    } /* switch count */
  } /* if ! blockmode */

}


/*
 * The function ioctl_dbcc() controls the behaviour of the driver
 */

static int 
ioctl_dbcc(struct inode * inode, struct file * file, 
	      unsigned cmd, unsigned long arg)
{

  unsigned long *ptr = (unsigned long *)arg;
  unsigned int crate = MINOR(inode->i_rdev);
  unsigned long temp;
  int result;

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  DEBUG(CAMAC_DBG_IOCTL,"camac: ioctl(crate=%u, n=%i, a=%i, cmd=%u)\n",
	crate, (int)N_NAF(file->f_pos),(int)A_NAF(file->f_pos),cmd);

  switch (cmd) {
  case CAMAC_INITIALIZE:
    dbcc_set(crate);
    dbcc_z_cycle();
    return 0;
  case CAMAC_CLEAR:
    dbcc_set(crate);
    dbcc_c_cycle();
    return 0;
  case CAMAC_C_Z_S2:
    dbcc_set(crate);
    dbcc_czs2();
    return 0;
  case CAMAC_TEST_CRATE:
    if (crate == AC_BROADCAST_CRATE) return -ENOIOCTLCMD;
    dbcc_set(crate);
    temp = dbcc_test();
    if (ptr != NULL) {
      if ((result = verify_area(VERIFY_WRITE,ptr,sizeof(long))))
	return result;
      memcpy_tofs(ptr, &temp, sizeof(long));
    }
    return 0;
  case CAMAC_STATUS:
    dbcc_set(crate);
    temp = dbcc_status();
    if (ptr != NULL) {
      if ((result = verify_area(VERIFY_WRITE,ptr,sizeof(long))))
	return result;
      memcpy_tofs(ptr, &temp, sizeof(long));
    }
    return 0;
  case CAMAC_BLOCK_MODE:
    ((cfg_t *)file->private_data)->Mode = CAMAC_BLOCK_MODE;
    return 0;
  case CAMAC_SINGLE_MODE:
    ((cfg_t *)file->private_data)->Mode = CAMAC_SINGLE_MODE;
    return 0;
  case CAMAC_ALLOW_BROADCAST:
    if (crate == AC_BROADCAST_CRATE) return -ENOIOCTLCMD;
    dbcc_set(crate);
    dbcc_allow_broadcast();
    return 0;
  case CAMAC_DENY_BROADCAST:
    if (crate == AC_BROADCAST_CRATE) return -ENOIOCTLCMD;
    dbcc_set(crate);
    dbcc_deny_broadcast();
    return 0;
  case CAMAC_TEST_LAM: /* they are here just for compability */
  case CAMAC_CLEAR_1:    
  case CAMAC_CLEAR_LAM:
  case CAMAC_CLEAR_2:
  case CAMAC_DISABLE:
  case CAMAC_EXECUTE:
  case CAMAC_ENABLE:
  case CAMAC_TEST_STATUS:
    dbcc_set(crate);
    temp = dbcc_ctrl(NAF(N_NAF(file->f_pos), A_NAF(file->f_pos),cmd) 
			| AC_START_CYCLE);
    if (ptr != NULL) {
      if ((result = verify_area(VERIFY_WRITE,ptr,sizeof(long))))
	return result;
      memcpy_tofs(ptr, &temp, sizeof(long));
    }
    return 0;
  case CAMAC_NON_DATA:
    dbcc_set(crate);
    temp = dbcc_ctrl(file->f_pos | AC_START_CYCLE);
    if (ptr != NULL) {
      if ((result = verify_area(VERIFY_WRITE,ptr,sizeof(long))))
	return result;
      memcpy_tofs(ptr, &temp, sizeof(long));
    }
    return 0;
#ifdef WAIT_IN_READ
  case CAMAC_SET_LAM:
    if ((result = verify_area(VERIFY_READ,ptr,sizeof(long))))
      return result;
    ((cfg_t *)file->private_data)->TimeOut = get_fs_long(arg);
    if (((cfg_t *)file->private_data)->TimeOut > AC_MAX_TIMEOUT)
      ((cfg_t *)file->private_data)->TimeOut = AC_MAX_TIMEOUT;
    return 0;
  case CAMAC_NOWAIT_LAM:
    ((cfg_t *)file->private_data)->TimeOut = -1;
    return 0;
  case CAMAC_WAIT_LAM:
    if (dbcc_wait_lam(file->f_pos, ((cfg_t *)file->private_data)->TimeOut))
      return -EIO;
    return 0;
#else /* !WAIT_IN_READ */
  case CAMAC_WAIT_LAM:
    if ((result = verify_area(VERIFY_READ,ptr,sizeof(long))))
      return result;
    ((cfg_t *)file->private_data)->TimeOut = get_fs_long(arg);
    if (((cfg_t *)file->private_data)->TimeOut > AC_MAX_TIMEOUT)
      ((cfg_t *)file->private_data)->TimeOut = AC_MAX_TIMEOUT;
    if (dbcc_wait_lam(file->f_pos, ((cfg_t *)file->private_data)->TimeOut))
      return -EIO;
    return 0;
#endif /* !WAIT_IN_READ */
  case CAMAC_SET_INHIBIT:  /* to be done */
    dbcc_set(crate);
    dbcc_set_inhibit();
    return 0;
  case CAMAC_CLEAR_INHIBIT: /* too be done */
    dbcc_set(crate);
    dbcc_clear_inhibit();
    return 0;
  default: 
    return -ENOIOCTLCMD;
  }

}

/*
 * The function open_dbcc() tests the crate and opens it
 */

static int 
open_dbcc(struct inode * inode, struct file * file)
{

  unsigned int crate = MINOR(inode->i_rdev);

  if (crate > AC_MAX_CRATES)
    return -ENODEV;

  DEBUG(CAMAC_DBG_OPEN,"camac: open(crate=%u)\n", crate);

  /* 
   * Look if the crate was opened exclusively before 
   */

  if (ExclOpen[crate] != NULL) {
    DEBUG((CAMAC_DBG_OPEN|CAMAC_DBG_ERROR),"camac: open failed: Crate busy\n");
    return -EBUSY;
  }
  
  /*
   * If the crate is to be opened exclusively, it can't be opened before
   */

  if (((file->f_flags & O_EXCL)) && (nOpen[crate] != 0)){
    DEBUG((CAMAC_DBG_OPEN|CAMAC_DBG_ERROR),
	  "camac: exclusive open failed: Crate busy\n");
    return -EBUSY;
  }

  if (crate != AC_BROADCAST_CRATE) {
    if (!nOpen[crate]) {
      dbcc_set(crate);
      if (! dbcc_test()) {
	DEBUG((CAMAC_DBG_OPEN|CAMAC_DBG_ERROR),
	      "camac: open failed: Crate not found\n");
	return -ENODEV;
      } /* if !dbcc_test */
    } /* if !nOpen */
    Irq[crate] = dbcc_test_irq(); /* test interrupt capability */
    
    if (Irq[crate]) {
      DEBUG(CAMAC_DBG_OPEN,"camac: crate uses irq %i\n", Irq[crate]);
      dbcc_enable_interrupt();
    }

  } /* if crate != AC_BROADCAST_CRATE */

  /*
   * The priate_data field points to a structure which holds the modi
   * of the device
   */

  file->private_data = kmalloc(sizeof(cfg_t *), GFP_USER);
  ((cfg_t *)file->private_data)->Mode = CAMAC_SINGLE_MODE;
  ((cfg_t *)file->private_data)->TimeOut = -1; /* no time out -- direct poll */

  /* 
   * If the crate is opened exclusively, deny broadcast and mark 
   * it exclusively opened
   */

  if (file->f_flags & O_EXCL) {
    ExclOpen[crate] = file;
    if (crate != AC_BROADCAST_CRATE)  dbcc_deny_broadcast();
  }

  MOD_INC_USE_COUNT;
  nOpen[crate]++;

  return 0;

}


/*
 * The function close_dbcc() closes the device
 */

void 
close_dbcc(struct inode * inode, struct file * file)
{

  unsigned int crate = MINOR(inode->i_rdev);

  if (crate > AC_MAX_CRATES)
    return;

  DEBUG(CAMAC_DBG_OPEN,"camac: close(crate=%u)\n",crate);

  nOpen[crate]--;
  if (ExclOpen[crate] == file) {
    DEBUG(CAMAC_DBG_OPEN,"camac: close excl\n");
    ExclOpen[crate] = NULL;
  }

  if (crate != AC_BROADCAST_CRATE) {
    dbcc_set(crate);
    dbcc_allow_broadcast();
  }

  dbcc_set(crate);
  dbcc_allow_broadcast();
  if (Irq[crate]) 
    dbcc_disable_interrupt();

  kfree(file->private_data);

  MOD_DEC_USE_COUNT;

  return;

}


/*
 * The function init_module() is used by starting the module
 * The function dbcc_init() is used by Linux startup if the driver
 * is directly compiled in.
 */
#ifdef MODULE
#define INIT_FAIL -EIO
#define INIT_SUCCESS 0
#else
#define INIT_FAIL mem_start
#define INIT_SUCCESS mem_start
#endif

#ifdef MODULE
int 
init_module(void)
#else
unsigned long 
dbcc_init(unsigned long mem_start)
#endif
{
  int crate;
  int count = 0;

#ifndef MODULES
  if (!dbcc_enabled)
    return INIT_FAIL;
#endif

  if (register_chrdev(major, AC_NAME, &dbcc_fops)) {
    printk(" unable to get major %i for " AC_NAME " device\n",major);
    return INIT_FAIL;
  }

  printk(STARTUPSTRING);

  if (check_region(BaseAddress, AC_BASE_EXTEND)) {
    printk(" unable to access i/o address 0x%lx\n",BaseAddress);
    unregister_chrdev(major, AC_NAME);
    return INIT_FAIL;
  }

  if (! dbcc_test_pcboard()) {
    printk(" interface card not found at address 0x%lx\n",BaseAddress);
    unregister_chrdev(major, AC_NAME);
    return INIT_FAIL;
  }

  printk(" interface card at address 0x%lx\n",BaseAddress);
  request_region(BaseAddress, AC_BASE_EXTEND, AC_NAME);

  if(irq) 
    if (request_irq(irq, dbcc_irq_handler, SA_INTERRUPT, AC_NAME)) {
      printk(" unable to install irq %i. falling back to", irq);
      irq = 0;
    }

  if (irq)
    printk(" using irq %i\n",irq);
  else
    printk(" polling mode\n");

  currentDevice = -1; 
  for (crate = 0; crate < AC_MAX_CRATES; crate++) {
    
    nOpen[crate] = 0;
    ExclOpen[crate] = NULL;
    Sleeping[crate] = 0;
    wait_q[crate] = NULL;

    if (crate == AC_BROADCAST_CRATE)
      continue;                    /* no check for broadcasting crate */

    dbcc_set(crate);
    if (dbcc_test()) {
      count++;
      printk(" crate /dev/camac%i available", crate);
      if (irq) {
	if (dbcc_test_irq()) 
	  printk(" (irq)");
	else
	  printk(" (polling)");
      } /* if irq */
      printk("\n");
    } /* if dbcc_test() */
  } /* for crate */ 

  if (count == 0) 
    printk(" no crate found\n");

  return INIT_SUCCESS;

}

#ifdef MODULE
/*
 * cleanup_module() cleans up the module if it is not used
 * if compiled-in, there is ofcourse no cleanup possible
 */

void
cleanup_module(void)
{

  if (MOD_IN_USE) {
    printk(AC_NAME ": device busy, module not removed\n");
    return;
  }
  
  release_region(BaseAddress, AC_BASE_EXTEND);

  if (irq)
    free_irq(irq);

  if (unregister_chrdev(major, AC_NAME) != 0)
    printk(AC_NAME ": cleanup_module failed\n");
  else
    printk(AC_NAME ": module removed\n");
}

#endif /* def MODULE */

#ifndef MODULE
/*
 * The setup takes two (optional) parameters: the base address
 * and the interrupt number
 * type "dbcc=0x320,5" at LILO command line
 * for a loadable module, this can be done via 
 * "insmod dbcc.o base=0x320 irq=5"
 */
void dbcc_setup(char *str, int *ints)
{
  if (!strcmp(str, "off")) {
    dbcc_enabled = 0;
    return;
  }

  if (ints[0] > 0)
    BaseAddress=ints[1];

  if (ints[0] > 1)
    irq=ints[2];

  dbcc_enabled = 1;

}
#endif /* ifndef MODULE */

/*
 * Here starts the real I/O code. It was mostly stolen
 * from Thomas Mikolajski's ACCAMAC Turbo-Pascal library
 * He gave the permission to use it here.
 */

/*
 * dbcc_set() sets the current device if it is not already set
 */

static void
dbcc_set(unsigned crate)
{
  
  if (currentDevice != crate) {

    outw(0, CtrlOutAddress);
    outw((0x10 | (crate & 0x0f)), DataOutAddress);

    currentDevice = crate;
  } /* if currentDevice != crate */
}


/*
 * dbcc_test() tests the actual crate
 * and denies further broadcasting
 * 0 means: not found
 */

static int 
dbcc_test(void)
{

  int cbus;
  int present = 1;

  DEBUG(CAMAC_DBG_IO,"Test Crate %i\n",currentDevice);
  dbcc_deny_broadcast();
  outw_p(3, CtrlOutAddress); /* primary mode 3: read */
  cbus = inw_p(DataInAddress);
  if ((cbus & 0x04) == 0) {
    DEBUG(CAMAC_DBG_IO,"Deny broadcast test: failed\n");
    present = 0;
  }
  else {
    DEBUG(CAMAC_DBG_IO,"Deny broadcast test: passed\n");
  }

  dbcc_allow_broadcast();
  outw_p(3, CtrlOutAddress); /* primary mode 3: read */
  cbus = inw_p(DataInAddress);
  if ((cbus & 0x04) != 0) {
    DEBUG(CAMAC_DBG_IO,"Allow broadcast test: failed\n");
    present = 0;
  }
  else {
    DEBUG(CAMAC_DBG_IO,"Allow broadcast test: passed\n");
  }

#if 1 /* needed for Q and X response */
  outw_p(1, CtrlOutAddress); /* reset, clear crate */
  outw_p(6, DataOutAddress);
  outw_p(2, CtrlOutAddress); /* make c+z pulse */
#endif

  cbus = dbcc_status();
  if (!CAMAC_Q_FLAG(cbus) || !CAMAC_X_FLAG(cbus)) {
    DEBUG(CAMAC_DBG_IO,"CAMAC status test: failed\n");
    present = 0;
  }
  else {
    DEBUG(CAMAC_DBG_IO,"CAMAC status test: passed\n");
  }

  return present;
}

/*
 * dbcc_test_irq () tests the actual crate for interrupts
 * 0 means: no interrupt avilable
 * >0: use this interrupt
 */

static int 
dbcc_test_irq(void)
{
  int irqstatus = 1;

  if (!irq) 
    return 0;

  dbcc_enable_interrupt();
  outw_p(3, CtrlOutAddress); /* primary mode 3: read */
  if ((inw_p(DataInAddress) & 0x02) == 0)
    irqstatus = 0;

  dbcc_disable_interrupt();
  outw_p(3, CtrlOutAddress); /* primary mode 3: read */
  if ((inw_p(DataInAddress) & 0x02) != 0)
    irqstatus = 0;

  if (irqstatus)
    return irq;
  else
    return 0;

}

/*
 * dbcc_test_pcboard() tests the PC interface card
 * 0 means: not found
 */

static int 
dbcc_test_pcboard(void)
{
#if 0 /* is really a very ugly code!! */
   
  if (inw(DataInAddress) == 0xffff) /* quick and dirty -- may fail */
    return 0;
  else
    return 1;
#else
  return 1;
#endif
}

/*
 * dbcc_enable_interrupt() enables the interrupt for the
 * actual crate
 */
static void 
dbcc_enable_interrupt(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(12, DataOutAddress); /* submode 12: enable interrupt */
}

/*
 * dbcc_disable_interrupt() disables the interrupt for the
 * actual crate
 */
static void 
dbcc_disable_interrupt(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(13, DataOutAddress); /* submode 12: disable interrupt */
}

/*
 * dbcc_deny_broadcast() excludes the current crate from the 
 * broadcast operations
 */
static void 
dbcc_deny_broadcast(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(7, DataOutAddress); /* submode 7 */
  outw_p(2, CtrlOutAddress); /* primary mode 2 */
  outw_p(4, DataOutAddress); /* submode 4: broadcasting off */
}

/*
 * dbcc_allow_broadcast() allows  broadcast operations
 * at the current crate
 */

static void 
dbcc_allow_broadcast(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(7, DataOutAddress); /* submode 7 */
  outw_p(2, CtrlOutAddress); /* primary mode 2 */
  outw_p(0, DataOutAddress); /* submode 4: broadcasting on */
}


/*
 * dbcc_status() reads out the current status of the C and Q flag
 */

static int 
dbcc_status()
{
  return inw(CtrlInAddress);
}

/*
 * dbcc_ctrl() invokes the several control commands
 */

static int 
dbcc_ctrl(int naf)
{
  dbcc_load_naf(naf);
  return CAMAC_Q_FLAG(dbcc_status());
}

/*
 * dbcc_czs2() makes a C-Z-S2 cycle on the CAMAC dataway
 */

static void 
dbcc_czs2(void)
{
  dbcc_load_naf(0); 
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(6, DataOutAddress); /* submode 6 */
  outw_p(2, CtrlOutAddress); /* primary mode 2: exec CAMAC cycle with c&z */
}

/*
 * dbcc_c_cycle() makes a C cycle on the CAMAC dataway
 */

static void 
dbcc_c_cycle(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(8, DataOutAddress); /* submode 8: enable C */
  outw_p(6, DataOutAddress); /* submode 6: CAMAC cycle with c */
}

/*
 * dbcc_z_cycle() makes a Z cycle on the CAMAC dataway
 */

static void 
dbcc_z_cycle(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(9, DataOutAddress); /* submode 9: enable Z */
  outw_p(6, DataOutAddress); /* submode 6: CAMAC cycle with z */
}

/*
 * dbcc_set_inhibit() sets the CAMAC inhibit flag
 */

static void 
dbcc_set_inhibit(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(10, DataOutAddress); /* submode 10: set I flag */
}

/*
 * dbcc_clear_inhibit() clears the CAMAC inhibit flag
 */

static void 
dbcc_clear_inhibit(void)
{
  outw_p(1, CtrlOutAddress); /* primary mode 1 */
  outw_p(11, DataOutAddress); /* submode 11: clear I flag */
}

/*
 * The timeout function we need for the Look-at-me waiting.
 * (see below)
 */

static void 
dbcc_mark_timeout(unsigned long timed_out)
{
  *(int *)timed_out = 1;
}

/*
 * The IRQ handler is called if the interrupt is requested
 */

static void
dbcc_irq_handler(int interrupt, struct pt_regs *regs)
{
  int crate;

  /*
   * If the interrupt is not our --> just return
   */

  if (interrupt != irq)
    return;

  /*
   * Look if we really got an interrupt for an actual crate 
   * and position
   */

  for (crate = 0; crate < AC_MAX_CRATES; ++crate) {

    /*
     * if the device isn't open, or not sleeping: continue
     */

    if (nOpen[crate] == 0) continue;
    if (Sleeping[crate] == 0) continue;
    
    /*
     * Look that we really had a LAM in this crate
     */

    dbcc_set(crate);
    if (CAMAC_LAM_FLAG(dbcc_status()))
      wake_up(wait_q+crate);

  } /* for crate */

  return;

}

/*
 * dbcc_wait_lam() waits for the LAM-state if the timeout is >0
 * The timeout should be given in 10 milliseconds.
 * It returns zero on success and -EIO on error.
 */


static int
dbcc_wait_lam(int naf, int timeout)
{
  /*
   * if the timer<0, dont look for LAM, just return.
   */

  if (timeout <= 0)
    return 0;

  if (Irq[currentDevice])
    return dbcc_wait_lam_irq(naf, timeout);
  else
    return dbcc_wait_lam_polling(naf, timeout);

}

/*
 * For interrupt-capable crates we wait for the interrupt an then check
 * if we got really the irq for the actual crate and module position
 * (work in progress)
 */

static int 
dbcc_wait_lam_irq(int naf, int timeout)
{
  unsigned crate = currentDevice;

  /* 
   * set the timeout
   */

  current->timeout = jiffies + timeout;

  /*
   * Sleep until we got the interrupt
   */

  Sleeping[crate] = 1;
  interruptible_sleep_on(wait_q+crate);

  /*
   * Look if we ran into the timeout 
   */

  if (current->timeout == 0) {
    DEBUG(CAMAC_DBG_ERROR,"camac: Timeout waiting for LAM (c=%u, n=%i a=%i)\n",
	  crate, N_NAF(naf),A_NAF(naf));
    return -EIO; /* failure*/
  }
  else
    return 0;    /* success */
}

/*
 * If the Crate is not IRQ-capable, we must use the polling mode.
 * this is done with the CAMAC function 8 (Test LAM).
 */

static int 
dbcc_wait_lam_polling(int naf, int timeout)
{
  int q;
  unsigned crate;
  volatile int timed_out = 0;
  struct timer_list 
    timer_camac = {NULL, NULL, timeout, (int)&timed_out, dbcc_mark_timeout};

  /*
   * replace the Function code with the look-at-me code
   */

  naf = NAF(N_NAF(naf), A_NAF(naf), CAMAC_TEST_LAM) | AC_START_CYCLE;

  /*
   * Polling loop: q holds the Look-at-me flag.
   */

  q = dbcc_ctrl(naf);
  add_timer(&timer_camac);

  /*
   * In the loop, we will allow a re-scheduling, so that other
   * processes may use another crate and change the current crate.
   * So, we must store here the current crate locally, and restore it
   * if nessessary.
   */

  crate = currentDevice;

  while ((!timed_out) && (q == 0)) {
    if (need_resched) {
      schedule();
      dbcc_set(crate);
    }
    q = dbcc_ctrl(naf);
  }

  del_timer(&timer_camac);
    
  if (timed_out) {
    DEBUG(CAMAC_DBG_ERROR,"camac: Timeout polling LAM (c=%u, n=%i a=%i)\n",
	  crate, N_NAF(naf),A_NAF(naf));
    return -EIO; /* failure*/
  }
  else
    return 0;    /* success */
}


/*
 * dbcc_load_naf() is the first part of the CAMAC cycle
 * it writes the Device number (n), the subaddress (a) and the
 * function code (f) to the interface. The NAF code can be calculated 
 * with the macro NAF(n, a, f)
 */

static void 
dbcc_load_naf(int naf)
{

  DEBUG(CAMAC_DBG_IO,"camac: dbcc_load_NAF(%i,%i,%i)\n",
	N_NAF(naf),A_NAF(naf),F_NAF(naf));

  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(0, DataOutAddress); /* submode 0 */
  outw(2, CtrlOutAddress); /* primary mode 2 */
  outw(naf, DataOutAddress); /* load naf */
}


/*
 * dbcc_read_16() is the second part of the CAMAC cycle; reads 16 bit
 */

static int 
dbcc_read_16(void)
{
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(3, DataOutAddress); /* submode 3 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */

  return inw(DataInAddress); /* read data */
}

/*
 * dbcc_read_24() is the second part of the CAMAC cycle; reads 24 bit
 */

static int 
dbcc_read_24(void)
{
  int data;

  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(3, DataOutAddress); /* submode 3 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */
  data = inw(DataInAddress); /* read lower 16 bit data */
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(4, DataOutAddress); /* submode 4 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */

  return ((inw(DataInAddress) & 0xff) << 16) + (data & 0xffff);/* high 8 bit */
}

/*
 * dbcc_read_blk() is the second part of the CAMAC cycle; 
 * reads in block mode
 */

static void
dbcc_read_blk(short *data, int nwords)
{
  int i;
  
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(5, DataOutAddress); /* submode 5 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */

  for (i = 0; i < nwords-1; i++)
    data[i] = inw(DataInAddress);       /* read the data (16 bit) */

  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(3, DataOutAddress); /* submode 3 */
  outw(3, CtrlOutAddress); /* primary mode 3: read */
  data[i] = inw(DataInAddress); /* read the last data word */

}


/*
 * dbcc_write_16() is the second part of the CAMAC cycle; writes 16 bit
 */

static void
dbcc_write_16(int data)
{

  DEBUG(CAMAC_DBG_IO,"camac: dbcc_write_16(0x%x)\n",data);

  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(1, DataOutAddress); /* submode 1 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw(0x0000, DataOutAddress); /* write high order bits: 0 */
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(2, DataOutAddress); /* submode 2 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw((data & 0xffff), DataOutAddress); /* write low order bits */
}

/*
 * dbcc_write_24() is the second part of the CAMAC cycle; writes 24 bit
 */

static void
dbcc_write_24(int data)
{
  DEBUG(CAMAC_DBG_IO,"camac: dbcc_write_24(0x%x)\n",data);

  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(1, DataOutAddress); /* submode 1 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw(((data & 0xff0000) >> 16), DataOutAddress); /* write high order bits */
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(2, DataOutAddress); /* submode 2 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  outw((data & 0xffff), DataOutAddress); /* write low order bits */
}

/*
 * dbcc_write_blk() is the second part of the CAMAC cycle; 
 * writes in block mode
 */

static void
dbcc_write_blk(short *data, int nwords)
{
  int i;
  
  outw(1, CtrlOutAddress); /* primary mode 1 */
  outw(2, DataOutAddress); /* submode 2 */
  outw(2, CtrlOutAddress); /* primary mode 2: write */
  for (i = 0; i < nwords; i++)
    outw(data[i], DataOutAddress); /* write data bits */

  DEBUG(CAMAC_DBG_IO,"camac: dbcc_write_blk(*data, %i)\n",nwords);
}

/*
 * Here this driver code ends. Good luck!
 *
 * Tschuessi
 *
 * Ole
 */


